@@ -5,6 +5,7 @@
 #include "push_util.h"
 #include "refspec.h"
 #include "remote.h"
+#include "futils.h"
 
 static git_repository *_repo;
 
@@ -20,7 +21,12 @@ static char *_remote_ssh_passphrase = NULL;
 static char *_remote_default = NULL;
 static char *_remote_expectcontinue = NULL;
 
-static int cred_acquire_cb(git_credential **,	const char *, const char *, unsigned int, void *);
+static char *_remote_push_options = NULL;
+
+static char *_orig_ssh_cmd = NULL;
+static char *_ssh_cmd = NULL;
+
+static int cred_acquire_cb(git_credential **, const char *, const char *, unsigned int, void *);
 
 static git_remote *_remote;
 static record_callbacks_data _record_cbs_data = {{ 0 }};
@@ -367,8 +373,17 @@ void test_online_push__initialize(void)
 	_remote_ssh_passphrase = cl_getenv("GITTEST_REMOTE_SSH_PASSPHRASE");
 	_remote_default = cl_getenv("GITTEST_REMOTE_DEFAULT");
 	_remote_expectcontinue = cl_getenv("GITTEST_REMOTE_EXPECTCONTINUE");
+	_remote_push_options = cl_getenv("GITTEST_PUSH_OPTIONS");
 	_remote = NULL;
 
+	_orig_ssh_cmd = cl_getenv("GIT_SSH_COMMAND");
+	_ssh_cmd = cl_getenv("GITTEST_SSH_CMD");
+
+	if (_ssh_cmd)
+		cl_setenv("GIT_SSH_COMMAND", _ssh_cmd);
+	else
+		cl_setenv("GIT_SSH_COMMAND", NULL);
+
 	/* Skip the test if we're missing the remote URL */
 	if (!_remote_url)
 		cl_skip();
@@ -422,6 +437,11 @@ void test_online_push__cleanup(void)
 	git__free(_remote_ssh_passphrase);
 	git__free(_remote_default);
 	git__free(_remote_expectcontinue);
+	git__free(_remote_push_options);
+
+	cl_setenv("GIT_SSH_COMMAND", _orig_ssh_cmd);
+	git__free(_orig_ssh_cmd);
+	git__free(_ssh_cmd);
 
 	/* Freed by cl_git_sandbox_cleanup */
 	_repo = NULL;
@@ -430,6 +450,7 @@ void test_online_push__cleanup(void)
 
 	record_callbacks_data_clear(&_record_cbs_data);
 
+	cl_fixture_cleanup("push-options-result");
 	cl_fixture_cleanup("testrepo.git");
 	cl_git_sandbox_cleanup();
 }
@@ -472,7 +493,8 @@ static void do_push(
 	const char *refspecs[], size_t refspecs_len,
 	push_status expected_statuses[], size_t expected_statuses_len,
 	expected_ref expected_refs[], size_t expected_refs_len,
-	int expected_ret, int check_progress_cb, int check_update_tips_cb)
+	int expected_ret, int check_progress_cb, int check_update_tips_cb,
+	git_strarray *remote_push_options)
 {
 	git_push_options opts = GIT_PUSH_OPTIONS_INIT;
 	size_t i;
@@ -484,6 +506,9 @@ static void do_push(
 		/* Auto-detect the number of threads to use */
 		opts.pb_parallelism = 0;
 
+		if (remote_push_options)
+			memcpy(&opts.remote_push_options, remote_push_options, sizeof(git_strarray));
+
 		memcpy(&opts.callbacks, &_record_cbs, sizeof(git_remote_callbacks));
 		data = opts.callbacks.payload;
 
@@ -527,13 +552,12 @@ static void do_push(
 			verify_update_tips_callback(_remote, expected_refs, expected_refs_len);
 
 	}
-
 }
 
 /* Call push_finish() without ever calling git_push_add_refspec() */
 void test_online_push__noop(void)
 {
-	do_push(NULL, 0, NULL, 0, NULL, 0, 0, 0, 1);
+	do_push(NULL, 0, NULL, 0, NULL, 0, 0, 0, 1, NULL);
 }
 
 void test_online_push__b1(void)
@@ -543,7 +567,9 @@ void test_online_push__b1(void)
 	expected_ref exp_refs[] = { { "refs/heads/b1", &_oid_b1 } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__b2(void)
@@ -553,7 +579,9 @@ void test_online_push__b2(void)
 	expected_ref exp_refs[] = { { "refs/heads/b2", &_oid_b2 } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__b3(void)
@@ -563,7 +591,9 @@ void test_online_push__b3(void)
 	expected_ref exp_refs[] = { { "refs/heads/b3", &_oid_b3 } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__b4(void)
@@ -573,7 +603,9 @@ void test_online_push__b4(void)
 	expected_ref exp_refs[] = { { "refs/heads/b4", &_oid_b4 } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__b5(void)
@@ -583,13 +615,15 @@ void test_online_push__b5(void)
 	expected_ref exp_refs[] = { { "refs/heads/b5", &_oid_b5 } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__b5_cancel(void)
 {
 	const char *specs[] = { "refs/heads/b5:refs/heads/b5" };
-	do_push(specs, ARRAY_SIZE(specs), NULL, 0, NULL, 0, GIT_EUSER, 1, 1);
+	do_push(specs, ARRAY_SIZE(specs), NULL, 0, NULL, 0, GIT_EUSER, 1, 1, NULL);
 }
 
 void test_online_push__multi(void)
@@ -620,7 +654,9 @@ void test_online_push__multi(void)
 	};
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 
 	cl_git_pass(git_reflog_read(&log, _repo, "refs/remotes/test/b1"));
 	entry = git_reflog_entry_byindex(log, 0);
@@ -641,16 +677,21 @@ void test_online_push__implicit_tgt(void)
 	const char *specs2[] = { "refs/heads/b2" };
 	push_status exp_stats2[] = { { "refs/heads/b2", 1 } };
 	expected_ref exp_refs2[] = {
-	{ "refs/heads/b1", &_oid_b1 },
-	{ "refs/heads/b2", &_oid_b2 }
+		{ "refs/heads/b1", &_oid_b1 },
+		{ "refs/heads/b2", &_oid_b2 }
 	};
 
 	do_push(specs1, ARRAY_SIZE(specs1),
 		exp_stats1, ARRAY_SIZE(exp_stats1),
-		exp_refs1, ARRAY_SIZE(exp_refs1), 0, 1, 1);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		0, 1, 1,
+		NULL);
+
 	do_push(specs2, ARRAY_SIZE(specs2),
 		exp_stats2, ARRAY_SIZE(exp_stats2),
-		exp_refs2, ARRAY_SIZE(exp_refs2), 0, 0, 0);
+		exp_refs2, ARRAY_SIZE(exp_refs2),
+		0, 0, 0,
+		NULL);
 }
 
 void test_online_push__fast_fwd(void)
@@ -672,19 +713,27 @@ void test_online_push__fast_fwd(void)
 
 	do_push(specs_init, ARRAY_SIZE(specs_init),
 		exp_stats_init, ARRAY_SIZE(exp_stats_init),
-		exp_refs_init, ARRAY_SIZE(exp_refs_init), 0, 1, 1);
+		exp_refs_init, ARRAY_SIZE(exp_refs_init),
+		0, 1, 1,
+		NULL);
 
 	do_push(specs_ff, ARRAY_SIZE(specs_ff),
 		exp_stats_ff, ARRAY_SIZE(exp_stats_ff),
-		exp_refs_ff, ARRAY_SIZE(exp_refs_ff), 0, 0, 0);
+		exp_refs_ff, ARRAY_SIZE(exp_refs_ff),
+		0, 0, 0,
+		NULL);
 
 	do_push(specs_reset, ARRAY_SIZE(specs_reset),
 		exp_stats_init, ARRAY_SIZE(exp_stats_init),
-		exp_refs_init, ARRAY_SIZE(exp_refs_init), 0, 0, 0);
+		exp_refs_init, ARRAY_SIZE(exp_refs_init),
+		0, 0, 0,
+		NULL);
 
 	do_push(specs_ff_force, ARRAY_SIZE(specs_ff_force),
 		exp_stats_ff, ARRAY_SIZE(exp_stats_ff),
-		exp_refs_ff, ARRAY_SIZE(exp_refs_ff), 0, 0, 0);
+		exp_refs_ff, ARRAY_SIZE(exp_refs_ff),
+		0, 0, 0,
+		NULL);
 }
 
 void test_online_push__tag_commit(void)
@@ -694,7 +743,9 @@ void test_online_push__tag_commit(void)
 	expected_ref exp_refs[] = { { "refs/tags/tag-commit", &_tag_commit } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__tag_tree(void)
@@ -704,7 +755,9 @@ void test_online_push__tag_tree(void)
 	expected_ref exp_refs[] = { { "refs/tags/tag-tree", &_tag_tree } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__tag_blob(void)
@@ -714,7 +767,9 @@ void test_online_push__tag_blob(void)
 	expected_ref exp_refs[] = { { "refs/tags/tag-blob", &_tag_blob } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__tag_lightweight(void)
@@ -724,7 +779,9 @@ void test_online_push__tag_lightweight(void)
 	expected_ref exp_refs[] = { { "refs/tags/tag-lightweight", &_tag_lightweight } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__tag_to_tag(void)
@@ -734,7 +791,9 @@ void test_online_push__tag_to_tag(void)
 	expected_ref exp_refs[] = { { "refs/tags/tag-tag", &_tag_tag } };
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 0, 0);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 0, 0,
+		NULL);
 }
 
 void test_online_push__force(void)
@@ -751,17 +810,80 @@ void test_online_push__force(void)
 
 	do_push(specs1, ARRAY_SIZE(specs1),
 		exp_stats1, ARRAY_SIZE(exp_stats1),
-		exp_refs1, ARRAY_SIZE(exp_refs1), 0, 1, 1);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		0, 1, 1,
+		NULL);
 
 	do_push(specs2, ARRAY_SIZE(specs2),
 		NULL, 0,
-		exp_refs1, ARRAY_SIZE(exp_refs1), GIT_ENONFASTFORWARD, 0, 0);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		GIT_ENONFASTFORWARD, 0, 0,
+		NULL);
 
 	/* Non-fast-forward update with force should pass. */
 	record_callbacks_data_clear(&_record_cbs_data);
 	do_push(specs2_force, ARRAY_SIZE(specs2_force),
 		exp_stats2_force, ARRAY_SIZE(exp_stats2_force),
-		exp_refs2_force, ARRAY_SIZE(exp_refs2_force), 0, 1, 1);
+		exp_refs2_force, ARRAY_SIZE(exp_refs2_force),
+		0, 1, 1,
+		NULL);
+}
+
+static void push_option_test(git_strarray given_options, const char *expected_option)
+{
+	const char *specs[] = { "refs/heads/b1:refs/heads/b1" };
+	push_status exp_stats[] = { { "refs/heads/b1", 1 } };
+	expected_ref exp_refs[] = { { "refs/heads/b1", &_oid_b1 } };
+	git_str push_options_path = GIT_STR_INIT;
+	git_str push_options_result = GIT_STR_INIT;
+	char *options[16];
+	git_strarray push_options = { options, given_options.count + 1 };
+	size_t i;
+
+	/* Skip the test if we're missing the push options result file */
+	if (!_remote_push_options)
+		cl_skip();
+
+	cl_assert(given_options.count < 16);
+
+	cl_git_pass(git_str_joinpath(&push_options_path, clar_sandbox_path(), "push-options-result"));
+
+	options[0] = push_options_path.ptr;
+	for (i = 0; i < given_options.count; i++)
+		options[i + 1] = given_options.strings[i];
+
+	do_push(specs, ARRAY_SIZE(specs),
+		exp_stats, ARRAY_SIZE(exp_stats),
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		&push_options);
+
+	cl_assert(git_fs_path_exists(push_options_path.ptr));
+	cl_git_pass(git_futils_readbuffer(&push_options_result, push_options_path.ptr));
+
+	cl_assert_equal_s(expected_option, git_str_cstr(&push_options_result));
+	git_str_dispose(&push_options_result);
+	git_str_dispose(&push_options_path);
+}
+
+void test_online_push__options(void)
+{
+	char *push_options_string_args_test_1[1] = { "test_string" };
+	git_strarray push_options_test_1 = { push_options_string_args_test_1, 1 };
+
+	char *push_options_string_args_test_2[2] = { "test_string", "another arg?" };
+	git_strarray push_options_test_2 = { push_options_string_args_test_2, 2 };
+
+	char *push_options_string_args_test_3[1] = { "👨🏿‍💻 but can it do unicode? 🇺🇦" };
+	git_strarray push_options_test_3 = { push_options_string_args_test_3, 1 };
+
+	char *push_options_string_args_test_4[3] = { "\0", "\0", "\0" };
+	git_strarray push_options_test_4 = { push_options_string_args_test_4, 3 };
+
+	push_option_test(push_options_test_1, "test_string");
+	push_option_test(push_options_test_2, "test_stringanother arg?");
+	push_option_test(push_options_test_3, "👨🏿‍💻 but can it do unicode? 🇺🇦");
+	push_option_test(push_options_test_4, "\0\0\0");
 }
 
 void test_online_push__delete(void)
@@ -792,7 +914,9 @@ void test_online_push__delete(void)
 
 	do_push(specs1, ARRAY_SIZE(specs1),
 		exp_stats1, ARRAY_SIZE(exp_stats1),
-		exp_refs1, ARRAY_SIZE(exp_refs1), 0, 1, 1);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		0, 1, 1,
+		NULL);
 
 	/* When deleting a non-existent branch, the git client sends zero for both
 	 * the old and new commit id.  This should succeed on the server with the
@@ -802,23 +926,35 @@ void test_online_push__delete(void)
 	 */
 	do_push(specs_del_fake, ARRAY_SIZE(specs_del_fake),
 		exp_stats_fake, 1,
-		exp_refs1, ARRAY_SIZE(exp_refs1), 0, 0, 0);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		0, 0, 0,
+		NULL);
+
 	do_push(specs_del_fake_force, ARRAY_SIZE(specs_del_fake_force),
 		exp_stats_fake, 1,
-		exp_refs1, ARRAY_SIZE(exp_refs1), 0, 0, 0);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		0, 0, 0,
+		NULL);
 
 	/* Delete one of the pushed branches. */
 	do_push(specs_delete, ARRAY_SIZE(specs_delete),
 		exp_stats_delete, ARRAY_SIZE(exp_stats_delete),
-		exp_refs_delete, ARRAY_SIZE(exp_refs_delete), 0, 0, 0);
+		exp_refs_delete, ARRAY_SIZE(exp_refs_delete),
+		0, 0, 0,
+		NULL);
 
 	/* Re-push branches and retry delete with force. */
 	do_push(specs1, ARRAY_SIZE(specs1),
 		exp_stats1, ARRAY_SIZE(exp_stats1),
-		exp_refs1, ARRAY_SIZE(exp_refs1), 0, 0, 0);
+		exp_refs1, ARRAY_SIZE(exp_refs1),
+		0, 0, 0,
+		NULL);
+
 	do_push(specs_delete_force, ARRAY_SIZE(specs_delete_force),
 		exp_stats_delete, ARRAY_SIZE(exp_stats_delete),
-		exp_refs_delete, ARRAY_SIZE(exp_refs_delete), 0, 0, 0);
+		exp_refs_delete, ARRAY_SIZE(exp_refs_delete),
+		0, 0, 0,
+		NULL);
 }
 
 void test_online_push__bad_refspecs(void)
@@ -862,7 +998,9 @@ void test_online_push__expressions(void)
 
 	do_push(specs_left_expr, ARRAY_SIZE(specs_left_expr),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 }
 
 void test_online_push__notes(void)
@@ -884,13 +1022,17 @@ void test_online_push__notes(void)
 
 	do_push(specs, ARRAY_SIZE(specs),
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 
 	/* And make sure to delete the note */
 
 	do_push(specs_del, ARRAY_SIZE(specs_del),
 		exp_stats, 1,
-		NULL, 0, 0, 0, 0);
+		NULL, 0,
+		0, 0, 0,
+		NULL);
 
 	git_signature_free(signature);
 }
@@ -920,13 +1062,17 @@ void test_online_push__configured(void)
 
 	do_push(NULL, 0,
 		exp_stats, ARRAY_SIZE(exp_stats),
-		exp_refs, ARRAY_SIZE(exp_refs), 0, 1, 1);
+		exp_refs, ARRAY_SIZE(exp_refs),
+		0, 1, 1,
+		NULL);
 
 	/* And make sure to delete the note */
 
 	do_push(specs_del, ARRAY_SIZE(specs_del),
 		exp_stats, 1,
-		NULL, 0, 0, 0, 0);
+		NULL, 0,
+		0, 0, 0,
+		NULL);
 
 	git_signature_free(signature);
 }
